@@ -1,8 +1,8 @@
 /* tifffastcrop
 
- v. 1.3.10
+ v. 1.4
 
- Copyright (c) 2013-2017 Christophe Deroulers
+ Copyright (c) 2013-2021 Christophe Deroulers
 
  Portions are based on libtiff's tiffcp code. tiffcp is:
  Copyright (c) 1988-1997 Sam Leffler
@@ -17,6 +17,7 @@
 #include <stdlib.h> /* exit, strtoul */
 #include <string.h>
 #include <strings.h> /* strncasecmp */
+#include <ctype.h> /* isdigit */
 #include <assert.h>
 #include <errno.h>
 #include <tiff.h>
@@ -35,12 +36,13 @@
 
 #define JPEG_MAX_DIMENSION 65500L /* in libjpeg's jmorecfg.h */
 
-#define EXIT_SYNTAX_ERROR        1
-#define EXIT_IO_ERROR            2
-#define EXIT_UNHANDLED_FILE_TYPE 3
-#define EXIT_INSUFFICIENT_MEMORY 4
+#define EXIT_SYNTAX_ERROR         1
+#define EXIT_IO_ERROR             2
+#define EXIT_UNHANDLED_INPUT_IMAGE_TYPE 3
+#define EXIT_INSUFFICIENT_MEMORY  4
 #define EXIT_UNABLE_TO_ACHIEVE_TILE_DIMENSIONS 5
-#define EXIT_GEOMETRY_ERROR	6
+#define EXIT_GEOMETRY_ERROR	  6
+#define EXIT_UNHANDLED_OUTPUT_FILE_TYPE 7
 
 #define CopyField(tag, v) \
     if (TIFFGetField(in, tag, &v)) TIFFSetField(TIFFout, tag, v)
@@ -53,6 +55,10 @@ static uint32 requestedxmin = 0;
 static uint32 requestedymin = 0;
 static uint32 requestedwidth = 0;
 static uint32 requestedlength = 0;
+static uint64 diroff = 0;
+static uint16 number_of_dirnum_ranges = 0;
+static uint16 * dirnum_ranges_starts = NULL;
+static uint16 * dirnum_ranges_ends = NULL;
 static int verbose = 0;
 
 #define OUTPUT_FORMAT_TIFF 0
@@ -161,6 +167,21 @@ static const char * searchSuffix(const char * path)
 }
 
 
+static int shouldBeHandled(uint16 dirnum)
+{
+	if (number_of_dirnum_ranges == 0)
+		return 1;
+
+	uint16 rn;
+	for (rn= 0 ; rn < number_of_dirnum_ranges ; rn++)
+		if (dirnum_ranges_starts[rn] <= dirnum &&
+		    dirnum_ranges_ends[rn] >= dirnum)
+			return 1;
+
+	return 0;
+}
+
+
 static void tiffCopyFieldsButDimensions(TIFF* in, TIFF* TIFFout)
 {
 	uint16 bitspersample, samplesperpixel, compression, shortv, *shortav;
@@ -446,8 +467,15 @@ static int testAndFixOutTIFFPhotoAndCompressionParameters(TIFF* in,
 		    spp == 1 ? PHOTOMETRIC_LOGL : PHOTOMETRIC_LOGLUV);
 	else if (compression == COMPRESSION_LZW ||
 		    compression == COMPRESSION_ADOBE_DEFLATE ||
-		    compression == COMPRESSION_DEFLATE/* ||
-	    	compression == COMPRESSION_LZMA*/) {
+		    compression == COMPRESSION_DEFLATE
+		    /* || compression == COMPRESSION_LZMA*/
+#ifdef HAVE_ZSTD
+		    || compression == COMPRESSION_ZSTD
+#endif
+#ifdef HAVE_WEBP
+		    || compression == COMPRESSION_WEBP
+#endif
+                        ) {
 		if (defpredictor != (uint16)-1)
 			TIFFSetField(TIFFout, TIFFTAG_PREDICTOR,
 				defpredictor);
@@ -692,12 +720,12 @@ static size_t computeMemorySize(uint16 spp, uint16 bitspersample,
 }
 
 
-static int makeExtractFromTIFFFile(const char * infilename,
-	const char * outfilename)
+static int makeExtractFromTIFFDirectory(const char * infilename, TIFF * in,
+        uint64 diroff, uint16 dirnum, uint16 numberdirs,
+        const char * outfilename)
 {
-	TIFF * in;
 	uint32 inimagewidth, inimagelength;
-	uint32 outwidth = requestedwidth, outlength = requestedlength;
+	uint32 outwidth = 0, outlength = 0;
 	uint16 planarconfig, spp, bitspersample;
 	size_t outmemorysize;
 	char * ouroutfilename = NULL;
@@ -706,22 +734,6 @@ static int makeExtractFromTIFFFile(const char * infilename,
 	uint32 y_of_last_read_scanline = 0;
 	int return_code = 0; /* Success */
 
-	if (requestedwidth == 0 || requestedlength == 0) {
-		fprintf(stderr, "Requested extract is empty. Can't do it. Aborting.\n");
-		return EXIT_GEOMETRY_ERROR;
-	}
-
-	in = TIFFOpen(infilename, "r");
-	if (in == NULL) {
-		if (verbose)
-			fprintf(stderr, "Unable to open file \"%s\".\n",
-				infilename);
-		return EXIT_IO_ERROR;
-	}
-
-	if (verbose)
-		fprintf(stderr, "File \"%s\" open.\n", infilename);
-
 	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &inimagewidth);
 	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &inimagelength);
 	TIFFGetFieldDefaulted(in, TIFFTAG_SAMPLESPERPIXEL, &spp);
@@ -730,19 +742,18 @@ static int makeExtractFromTIFFFile(const char * infilename,
 	if (output_format == OUTPUT_FORMAT_JPEG &&
 	    (bitspersample != 8 || spp != 3)) {
 		TIFFError(TIFFFileName(in),
-			"Can't output JPEG file from file with "
+			"Can't output JPEG file from image with "
 			"bits-per-sample %u (not 8) or "
 			"samples-per-pixel %u (not 3). Aborting",
 			bitspersample, spp);
-		TIFFClose(in);
-		return EXIT_UNHANDLED_FILE_TYPE;
+		return EXIT_UNHANDLED_INPUT_IMAGE_TYPE;
 	} else if (output_format == OUTPUT_FORMAT_PNG) {
 		    /* We support only PNG_COLOR_TYPE_GRAY,
 		     PNG_COLOR_TYPE_RGB and PNG_COLOR_TYPE_RGB_ALPHA */
 		int success = 1;
 		if (spp != 1 && spp != 3 && spp != 4) {
 			TIFFError(TIFFFileName(in),
-				"Can't output PNG file from file with "
+				"Can't output PNG file from image with "
 				"samples-per-pixel %d (not 1, 3 or 4). Aborting",
 				spp);
 			success = 0;
@@ -752,7 +763,7 @@ static int makeExtractFromTIFFFile(const char * infilename,
 			    bitspersample != 4 && bitspersample != 8 &&
 			    bitspersample != 16) {
 				TIFFError(TIFFFileName(in),
-					"Error, can't output PNG file from file with 1 sample per pixel and "
+					"Error, can't output PNG file from image with 1 sample per pixel and "
 					"bitdepth (bits-per-sample) %u (not 1, 2, 4, 8 or 16). Aborting",
 					bitspersample);
 				success = 0;
@@ -763,7 +774,7 @@ static int makeExtractFromTIFFFile(const char * infilename,
 			case 4:
 			if (bitspersample != 8 && bitspersample != 16) {
 				TIFFError(TIFFFileName(in),
-					"Error, can't output PNG file from file with %u samples per pixel and "
+					"Error, can't output PNG file from image with %u samples per pixel and "
 					"bitdepth (bits-per-sample) %u (not 8 or 16). Aborting",
 					spp, bitspersample);
 				success = 0;
@@ -772,29 +783,34 @@ static int makeExtractFromTIFFFile(const char * infilename,
 
 			default:
 			fprintf(stderr, "Internal error.\n");
-			return EXIT_UNHANDLED_FILE_TYPE;
-		}
-		if (!success) {
-			TIFFClose(in);
-			return EXIT_UNHANDLED_FILE_TYPE;
+			return EXIT_UNHANDLED_INPUT_IMAGE_TYPE;
 		}
+		if (!success)
+			return EXIT_UNHANDLED_INPUT_IMAGE_TYPE;
 	} else if (bitspersample % 8 != 0 && 8 % bitspersample != 0) {
 		TIFFError(TIFFFileName(in),
-			"Error, can't deal with file with "
+			"Error, can't deal with image with "
 			"bits-per-sample %d (not a multiple nor a "
 			"divisor of 8)",
 			bitspersample);
-		TIFFClose(in);
-		return EXIT_UNHANDLED_FILE_TYPE;
+		return EXIT_UNHANDLED_INPUT_IMAGE_TYPE;
 	}
 
 	if (verbose)
-		fprintf(stderr, "File \"%s\" has %u bits per sample and %u samples per pixel.\n",
-			infilename, (unsigned) bitspersample, (unsigned) spp);
+		fprintf(stderr, "File \"%s\", directory %u has %u bits per sample and %u samples per pixel.\n",
+			infilename, dirnum, (unsigned) bitspersample, (unsigned) spp);
+
+	if (requestedwidth == (uint32) -1)
+		requestedwidth = inimagewidth - requestedxmin;
+	if (requestedlength == (uint32) -1)
+		requestedlength = inimagelength - requestedymin;
+	if (verbose)
+		fprintf(stderr, "Requested rectangle: " UINT32_FORMAT
+			"x" UINT32_FORMAT ".\n",
+			requestedwidth, requestedlength);
 
 	if (requestedxmin > inimagewidth || requestedymin > inimagelength) {
 		fprintf(stderr, "Requested top left corner is outside the image. Aborting.\n");
-		TIFFClose(in);
 		return EXIT_GEOMETRY_ERROR;
 	}
 	if (requestedxmin + requestedwidth > inimagewidth ||
@@ -814,38 +830,52 @@ static int makeExtractFromTIFFFile(const char * infilename,
 	TIFFGetField(in, TIFFTAG_PLANARCONFIG, &planarconfig);
 	if (planarconfig != PLANARCONFIG_CONTIG) {
 		TIFFError(TIFFFileName(in),
-			"Error, can't deal with file with "
+			"Error, can't deal with image with "
 			"planar configuration %d (non contiguous)",
 			planarconfig);
-		TIFFClose(in);
-		return EXIT_UNHANDLED_FILE_TYPE;
+		return EXIT_UNHANDLED_INPUT_IMAGE_TYPE;
 	}
 
 	if (output_format == OUTPUT_FORMAT_JPEG &&
 	    ( (requestedwidth >= JPEG_MAX_DIMENSION) ||
 	      (requestedlength >= JPEG_MAX_DIMENSION) ) ) {
 		fprintf(stderr, "At least one requested extract dimension is too large for JPEG files.\n");
-		TIFFClose(in);
 		return EXIT_UNABLE_TO_ACHIEVE_TILE_DIMENSIONS;
 	}
 
+	outwidth = requestedwidth; outlength = requestedlength;
 	outmemorysize= computeMemorySize(spp, bitspersample, outwidth,
 		outlength);
 	outbuf= outmemorysize == 0 ? NULL : malloc(outmemorysize);
 	if (outbuf == NULL) {
-		fprintf(stderr, "Unable to allocate enough memory to prepare extract ("
-		        UINT64_FORMAT " bytes needed).\n",
+		fprintf(stderr, "Unable to allocate enough memory to"
+			" prepare extract (%zu bytes needed).\n",
 		        outmemorysize);
-		TIFFClose(in);
 		return EXIT_INSUFFICIENT_MEMORY;
 	}
 
 	{
-	char * prefix = searchPrefixBeforeLastDot(infilename);
-	if (outfilename == NULL) {
+	char * prefix = searchPrefixBeforeLastDot(outfilename != NULL ?
+			    outfilename : infilename);
+	if (outfilename == NULL || diroff || numberdirs > 1) {
 		uint32 ndigitsx = searchNumberOfDigits(inimagewidth),
 		    ndigitsy = searchNumberOfDigits(inimagelength);
-		my_asprintf(&ouroutfilename, "%s-%0*u-%0*u-%0*ux%0*u.%s",
+		if (diroff)
+			my_asprintf(&ouroutfilename, "%s-d0x" UINT64_HEX_FORMAT "-%0*u-%0*u-%0*ux%0*u.%s",
+			    prefix, diroff, ndigitsx, requestedxmin,
+			    ndigitsy, requestedymin, ndigitsx,
+			    requestedwidth, ndigitsy, requestedlength,
+			    OUTPUT_SUFFIX[output_format]);
+		else if (numberdirs > 1) {
+			uint32 ndigitsdirnum = searchNumberOfDigits(numberdirs);
+			my_asprintf(&ouroutfilename, "%s-d%0*u-%0*u-%0*u-%0*ux%0*u.%s",
+			    prefix, ndigitsdirnum, dirnum,
+			    ndigitsx, requestedxmin,
+			    ndigitsy, requestedymin, ndigitsx,
+			    requestedwidth, ndigitsy, requestedlength,
+			    OUTPUT_SUFFIX[output_format]);
+		} else
+			my_asprintf(&ouroutfilename, "%s-%0*u-%0*u-%0*ux%0*u.%s",
 			    prefix, ndigitsx, requestedxmin, ndigitsy,
 			    requestedymin, ndigitsx,
 			    requestedwidth, ndigitsy, requestedlength,
@@ -872,10 +902,8 @@ static int makeExtractFromTIFFFile(const char * infilename,
 	}
 	if (ouroutfilename != NULL)
 		free(ouroutfilename);
-	if (out == NULL) {
-		TIFFClose(in);
+	if (out == NULL)
 		return EXIT_IO_ERROR;
-	}
 
 	{
 		uint16 in_compression;
@@ -943,7 +971,6 @@ static int makeExtractFromTIFFFile(const char * infilename,
 				fprintf(stderr, "Error, can't allocate space for row_pointers.\n");
 				fclose(out);
 				free(outbuf);
-				TIFFClose(in);
 				return EXIT_INSUFFICIENT_MEMORY;
 			}
 
@@ -1105,13 +1132,64 @@ static int makeExtractFromTIFFFile(const char * infilename,
 	default:
 		fprintf(stderr, "Unsupported output file format.\n");
 		free(outbuf);
-		TIFFClose(in);
-		return EXIT_UNHANDLED_FILE_TYPE;
+		return EXIT_UNHANDLED_OUTPUT_FILE_TYPE;
 	}
 
 	if (return_code == 0 && verbose)
 		fprintf(stderr, "Extract written.\n");
 	free(outbuf);
+	return return_code;
+}
+
+
+static int makeExtractFromTIFFFile(const char * infilename,
+	const char * outfilename)
+{
+	TIFF * in;
+	int return_code = 0; /* Success */
+
+	if (requestedwidth == 0 || requestedlength == 0) {
+		fprintf(stderr, "Requested extract is empty. Can't do it. Aborting.\n");
+		return EXIT_GEOMETRY_ERROR;
+	}
+
+	in = TIFFOpen(infilename, "r");
+	if (in == NULL) {
+		if (verbose)
+			fprintf(stderr, "Unable to open file \"%s\".\n",
+				infilename);
+		return EXIT_IO_ERROR;
+	}
+
+	if (verbose)
+		fprintf(stderr, "File \"%s\" open.\n", infilename);
+
+	if (diroff != 0) {
+		if (TIFFSetSubDirectory(in, diroff))
+			makeExtractFromTIFFDirectory(infilename, in,
+			    diroff, 0, 0, outfilename);
+	} else if (number_of_dirnum_ranges == 1 &&
+	    dirnum_ranges_starts[0] == 0 && dirnum_ranges_ends[0] == 0) {
+		 /* special case for speed (?) */
+		if (TIFFSetDirectory(in, (tdir_t) dirnum_ranges_starts[0]))
+			makeExtractFromTIFFDirectory(infilename, in,
+			    0, 0, 1, outfilename);
+	} else {
+		uint16 numberofdirectories= TIFFNumberOfDirectories(in);
+
+		if (verbose)
+			fprintf(stderr, "File \"%s\" has %u directories.\n",
+			    infilename, numberofdirectories);
+
+		uint16 curdir= 0;
+		do {
+			if (shouldBeHandled(curdir))
+				makeExtractFromTIFFDirectory(infilename, in,
+				    0, curdir, numberofdirectories,
+				    outfilename);
+			curdir++;
+		} while (TIFFReadDirectory(in));
+	}
 	TIFFClose(in);
 	return return_code;
 }
@@ -1119,7 +1197,7 @@ static int makeExtractFromTIFFFile(const char * infilename,
 
 static void usage()
 {
-	fprintf(stderr, "tifffastcrop v" PACKAGE_VERSION " license GNU GPL v3 (c) 2013-2017 Christophe Deroulers\n\n");
+	fprintf(stderr, "tifffastcrop v" PACKAGE_VERSION " license GNU GPL v3 (c) 2013-2021 Christophe Deroulers\n\n");
 	fprintf(stderr, "Quote \"Deroulers et al., Diagnostic Pathology 2013, 8:92\" in your production\n       http://doi.org/10.1186/1746-1596-8-92\n\n");
 	fprintf(stderr, "Usage: tifffastcrop [options] input.tif [output_name]\n\n");
 	fprintf(stderr, " Extracts (crops), without loading the full image input.tif into memory, a\nrectangular region from it, and saves it. Output file name is output_name if\ngiven, otherwise a name derived from input.tif. Output file format is guessed\nfrom output_name's extension if possible. Options:\n");
@@ -1127,17 +1205,20 @@ static void usage()
 	fprintf(stderr, " -T                report TIFF errors/warnings on stderr (no dialog boxes)\n");
 	fprintf(stderr, " -E x,y,w,l        region to extract/crop (x,y: coordinates of top left corner,\n");
 	fprintf(stderr, "   w,l: width and length in pixels)\n");
-        fprintf(stderr, " -j[#]             output JPEG file (with quality #, 0-100, default 75)\n");
+	fprintf(stderr, " -o offset         extracts only from directory at position offset in file\n");
+	fprintf(stderr, " -d range1[,range2...] extracts from dir. having numbers in the given ranges\n");
+	fprintf(stderr, "                   (numbers start at 0; ranges are like 3-3, 5:8, 4-, -0)\n");
+	fprintf(stderr, " -j[#]             output JPEG file (with quality #, 0-100, default 75)\n");
 #ifdef HAVE_PNG
-        fprintf(stderr, " -p[#]             output PNG file (with quality #, 0-9, default 6)\n");
+	fprintf(stderr, " -p[#]             output PNG file (with quality #, 0-9, default 6)\n");
 #endif
 	fprintf(stderr, " -c none[:opts]    output TIFF file with no compression \n");
-	fprintf(stderr, " -c x[:opts]       output TIFF compressed with encoding x (jpeg, lzw, zip, ...)\n");
+	fprintf(stderr, " -c x[:opts]       output TIFF compressed with encoding x (jpeg, lzw, zip,...)\n");
 	fprintf(stderr, "When output file format can't be guessed from the output filename extension, it is TIFF with same compression as input.\n\n");
 	fprintf(stderr, "JPEG-compressed TIFF options:\n");
 	fprintf(stderr, " #   set compression quality level (0-100, default 75)\n");
 /*	fprintf(stderr, " r  output color image as RGB rather than YCbCr\n");*/
-	fprintf(stderr, "LZW, Deflate (ZIP) and LZMA2 options:\n");
+	fprintf(stderr, "LZW, Deflate (ZIP), LZMA2, ZSTD and WEBP options:\n");
 	fprintf(stderr, " #   set predictor value\n");
 	fprintf(stderr, " p#  set compression level (preset)\n");
 	fprintf(stderr, "For example, -c lzw:2 to get LZW-encoded data with horizontal differencing,\n");
@@ -1245,6 +1326,18 @@ static int processCompressOptions(char* opt)
 		defcompression = COMPRESSION_JBIG;
 	} else if (strncasecmp(opt, "sgilog", 6) == 0) {
 		defcompression = COMPRESSION_SGILOG;
+#ifdef HAVE_ZSTD
+	} else if (strncasecmp(opt, "zstd", 4) == 0) {
+		if (!processZIPOptions(opt))
+			return 0;
+		defcompression = COMPRESSION_ZSTD;
+#endif
+#ifdef HAVE_WEBP
+	} else if (strncasecmp(opt, "webp", 4) == 0) {
+		if (!processZIPOptions(opt))
+			return 0;
+		defcompression = COMPRESSION_WEBP;
+#endif
 	} else
 		return (0);
 
@@ -1265,6 +1358,99 @@ static int processExtractGeometryOptions(char* cp)
 }
 
 
+ /* Add dirnum ranges to ranges that have already been stored */
+static unsigned parseDirnumRanges(const char* c, int dryrun)
+{
+	const char* c_depart = c;
+	uint16 number_of_read_ranges = 0;
+	uint16 debut_de_plage_courante = 0;
+	uint16 fin_de_plage_courante = (uint16) -1;
+	int separateur_debut_fin_vu = 0, premier_nombre_lu = 0,
+	    fin_de_la_chaine = 0;
+
+	do {
+		if (*c == 0)
+			fin_de_la_chaine = 1;
+		if (*c == ',' || *c == 0) {
+			c++;
+			if (premier_nombre_lu) {
+				if (! dryrun) {
+					dirnum_ranges_starts[number_of_dirnum_ranges + number_of_read_ranges] =
+					    debut_de_plage_courante;
+					dirnum_ranges_ends[number_of_dirnum_ranges + number_of_read_ranges]=
+					    separateur_debut_fin_vu
+					    ? fin_de_plage_courante /* range like "25:45" */
+					    : debut_de_plage_courante; /* range like "30" -> 30:30 */
+				}
+			number_of_read_ranges++;
+			}
+		debut_de_plage_courante = 0;
+		fin_de_plage_courante = (uint16) -1;
+		separateur_debut_fin_vu = 0;
+		premier_nombre_lu = 0;
+		continue;
+		}
+		if (*c == ':' || *c == '-') {
+			c++;
+			separateur_debut_fin_vu = 1;
+			continue;
+		}
+		if (isdigit(*c)) {
+			unsigned u;
+			int n;
+			if (sscanf(c, "%u%n", &u, &n) != 1) {
+				fprintf(stderr, "Syntax error in the list of directory number ranges \"%s\" at position \"%s\": integer number was expected.\n",
+				    c_depart, c);
+				return 0;
+			}
+			c += n;
+			if (separateur_debut_fin_vu)
+				fin_de_plage_courante = u;
+			else
+				debut_de_plage_courante = u;
+			premier_nombre_lu = 1;
+			continue;
+		}
+		fprintf(stderr, "Syntax error in the list of directory number ranges \"%s\" at position \"%s\": integer number was expected.\n",
+			c_depart, c);
+		return 0;
+	} while (! fin_de_la_chaine);
+
+	if (! dryrun)
+		number_of_dirnum_ranges += number_of_read_ranges;
+
+	return number_of_read_ranges;
+}
+
+ /* Add dirnum ranges to ranges that have already been stored (in case of
+   multiple lists of ranges on command line) */
+static int processDirnumRangesOption(const char* cp)
+{
+	uint16 number_of_read_ranges = parseDirnumRanges(cp, 1);
+
+	if (number_of_read_ranges == 0) {
+		fprintf(stderr, "Expected argument to -d option: non empty list of directory number ranges \"%s\".\n",
+		    cp);
+		return 0;
+	}
+
+	dirnum_ranges_starts = realloc(dirnum_ranges_starts,
+		sizeof(*dirnum_ranges_starts) *
+		(number_of_dirnum_ranges + number_of_read_ranges));
+	dirnum_ranges_ends = realloc(dirnum_ranges_ends,
+		sizeof(*dirnum_ranges_ends) *
+		(number_of_dirnum_ranges + number_of_read_ranges));
+	if (dirnum_ranges_starts == NULL || dirnum_ranges_ends == NULL) {
+		perror("Insufficient memory for directory number ranges ");
+		exit(EXIT_INSUFFICIENT_MEMORY);
+	}
+
+	parseDirnumRanges(cp, 0);
+
+	return 1;
+}
+
+
 static void stderrErrorHandler(const char* module, const char* fmt, va_list ap)
 {
 	if (module != NULL)
@@ -1309,6 +1495,27 @@ int main(int argc, char * argv[])
 			}
 			seen_extract_geometry_on_the_command_line = 1;
 			arg++;
+		} else if (argv[arg][1] == 'o') {
+			if (arg+1 >= argc) {
+				fprintf(stderr, "Option -o requires "
+					"an argument.\n");
+				usage();
+				return EXIT_SYNTAX_ERROR;
+			}
+			diroff = strtoul(argv[arg+1], NULL, 0);
+			arg++;
+		} else if (argv[arg][1] == 'd') {
+			if (arg+1 >= argc) {
+				fprintf(stderr, "Option -d requires "
+					"an argument.\n");
+				usage();
+				return EXIT_SYNTAX_ERROR;
+			}
+			if (!processDirnumRangesOption(argv[arg+1])) {
+				usage();
+				return EXIT_SYNTAX_ERROR;
+			}
+			arg++;
 		} else if (argv[arg][1] == 'c') {
 			output_format = OUTPUT_FORMAT_TIFF;
 			if (arg+1 >= argc ||
