@@ -1,8 +1,8 @@
 /* tiffmakemosaic
 
- v. 1.3.10
+ v. 1.4
 
- Copyright (c) 2012-2017 Christophe Deroulers
+ Copyright (c) 2012-2021 Christophe Deroulers
 
  Portions are based on libtiff's tiffcp code. tiffcp is
  Copyright (c) 1988-1997 Sam Leffler
@@ -42,8 +42,8 @@
 #define CopyField3(tag, v1, v2, v3) \
     if (TIFFGetField(in, tag, &v1, &v2, &v3)) TIFFSetField(out, tag, v1, v2, v3)
 
-static const char TIFF_SUFFIX[] = ".tif"; 
-static const char JPEG_SUFFIX[] = ".jpg"; 
+static const char TIFF_SUFFIX[] = ".tif";
+static const char JPEG_SUFFIX[] = ".jpg";
 static uint64 mosaicpiecesize = 1 << 30; /* 1 GiB */
 static uint32 overlapinpixels = 0;
 static long double overlapinpercent = 0;
@@ -52,6 +52,11 @@ static uint32 requestedpiecelength = 0;
 static uint32 requestedpiecewidthdivisor = 0;
 static uint32 requestedpiecelengthdivisor = 0;
 static int verbose = 0;
+static int dryrun = 0;
+static int paddinginx = 0;
+static int paddinginy = 0;
+static uint16 numberpaddingvalues = 0;
+static char ** paddingvalues = NULL;
 
 
 #ifndef TIFFMAKEMOSAIC_OUTPUTJPEGFILES
@@ -195,16 +200,35 @@ tiffCopyFieldsButDimensions(TIFF* in, TIFF* out)
 }
 
 
-static void cpBufToBuf(uint8* out, uint8* in, uint32 rows,
-	uint32 bytesperline, int outskew, int inskew)
+	/* cols resp. rows do not include rightpaddinginpixels resp. 
+	bottompadding */
+static void cpBufToBuf(uint8* out, uint8* in, uint8* paddingbytes,
+	uint32 rows, uint32 bottompadding, uint32 cols,
+	uint32 rightpaddinginpixels, uint16 bytesperpixel,
+	int outskew, int inskew)
 {
 	while (rows-- > 0) {
-		uint32 j = bytesperline;
-		while (j-- > 0)
+		uint32 jb = cols * bytesperpixel;
+		uint32 jp = rightpaddinginpixels;
+		while (jb-- > 0)
 			*out++ = *in++;
+		while (jp-- > 0) {
+			uint16 k;
+			for (k = 0 ; k < bytesperpixel ; k++)
+				*out++ = paddingbytes[k];
+		}
 		out += outskew;
 		in += inskew;
 	}
+	while (bottompadding-- > 0) {
+		uint32 jp = cols + rightpaddinginpixels;
+		while (jp-- > 0) {
+			uint32 k;
+			for (k = 0 ; k < bytesperpixel ; k++)
+				*out++ = paddingbytes[k];
+		}
+		out += outskew;
+	}
 }
 
 
@@ -238,10 +262,10 @@ static int testAndFixParameters(TIFF* in,
 		/* like in libtiff's tiffcp.c -- otherwise the reserved
 		 size for the tiles is too small and the program
 		 segfaults */
-		/* Force conversion to RGB */
+		/* force conversion to RGB */
 		TIFFSetField(in, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB);
 	} else if (input_photometric == PHOTOMETRIC_YCBCR) {
-		/* Otherwise, can't handle subsampled input */
+		/* otherwise, can't handle subsampled input */
 		uint16 subsamplinghor, subsamplingver;
 
 		TIFFGetFieldDefaulted(in, TIFFTAG_YCBCRSUBSAMPLING,
@@ -279,8 +303,15 @@ static int testAndFixParameters(TIFF* in,
 			    spp == 1 ? PHOTOMETRIC_LOGL : PHOTOMETRIC_LOGLUV);
 		else if (compression == COMPRESSION_LZW ||
 			    compression == COMPRESSION_ADOBE_DEFLATE ||
-			    compression == COMPRESSION_DEFLATE/* ||
-		    	compression == COMPRESSION_LZMA*/) {
+			    compression == COMPRESSION_DEFLATE
+			    /* || compression == COMPRESSION_LZMA*/
+#ifdef HAVE_ZSTD
+			    || compression == COMPRESSION_ZSTD
+#endif
+#ifdef HAVE_WEBP
+			    || compression == COMPRESSION_WEBP
+#endif
+                        ) {
 			if (defpredictor != (uint16)-1)
 				TIFFSetField(TIFFout, TIFFTAG_PREDICTOR,
 					defpredictor);
@@ -325,7 +356,7 @@ static int testAndFixParameters(TIFF* in,
 			free(pn);
 		}
 
-		/* To be done *after* setting compression -- otherwise,
+		/* to be done *after* setting compression -- otherwise,
 		 * ScanlineSize may be wrong */
 		*outscanlinesizeinbytes = TIFFScanlineSize(TIFFout);
 
@@ -338,10 +369,13 @@ static int testAndFixParameters(TIFF* in,
 	return 1;
 }
 
+	/* width resp. length include rightpadding resp. bottompadding */
 static int
 cpTiles2Strip(TIFF* in, void * ambiguous_out,
 	int output_to_jpeg_rather_than_tiff,
 	uint32 xmin, uint32 ymin, uint32 width, uint32 length,
+	uint32 rightpadding, uint32 bottompadding, uint8* paddingbytes,
+	uint32 inimagewidth, uint32 inimagelength,
 	unsigned char * outbuf)
 {
 	struct jpeg_compress_struct * p_cinfo;
@@ -382,17 +416,24 @@ cpTiles2Strip(TIFF* in, void * ambiguous_out,
 	for (y = ymin ; y < ymin + length + intilelength ; y += intilelength) {
 		uint32 x, colb = 0;
 		uint32 yminoftile = (y/intilelength) * intilelength;
-		uint32 ymintocopy = ymin > yminoftile ? ymin : yminoftile;
+		uint32 ymintocopyintile = ymin > yminoftile ? ymin : yminoftile;
 		uint32 ymaxplusone = yminoftile + intilelength;
-		uint32 lengthtocopy;
+		uint32 lengthtocopy, bottompaddingthistile = 0;
 		unsigned char * inbufrow = inbuf +
-		    intilewidthinbytes * (ymintocopy-yminoftile);
+		    intilewidthinbytes * (ymintocopyintile - yminoftile);
 
 		if (ymaxplusone > ymin + length)
 			ymaxplusone = ymin + length;
 		if (ymaxplusone <= yminoftile)
 			break;
-		lengthtocopy = ymaxplusone - ymintocopy;
+		if (yminoftile >= inimagelength) {
+			lengthtocopy = 0;
+			bottompaddingthistile = ymaxplusone - yminoftile;
+		} else if (ymaxplusone > inimagelength) {
+			lengthtocopy = inimagelength - ymintocopyintile;
+			bottompaddingthistile = ymaxplusone - inimagelength;
+		} else
+			lengthtocopy = ymaxplusone - ymintocopyintile;
 
 		for (x = xmin ; x < xmin + width + intilewidth ;
 		    x += intilewidth) {
@@ -400,16 +441,27 @@ cpTiles2Strip(TIFF* in, void * ambiguous_out,
 			uint32 xmintocopyintile = xmin > xminoftile ?
 			    xmin : xminoftile;
 			uint32 xmaxplusone = xminoftile + intilewidth;
-			tsize_t widthtocopyinbytes;
+			uint32 widthtocopy, rightpaddingthistile = 0;
+			tsize_t widthtoreadinbytes, widthtowriteinbytes;
 
 			if (xmaxplusone > xmin + width)
 				xmaxplusone = xmin + width;
 			if (xmaxplusone <= xminoftile)
 				break;
-			widthtocopyinbytes = (xmaxplusone - xmintocopyintile) *
-				bytesperpixel;
+			if (xminoftile >= inimagewidth) {
+				widthtocopy = 0;
+				rightpaddingthistile = xmaxplusone - xminoftile;
+			} else if (xmaxplusone > inimagewidth) {
+				widthtocopy = inimagewidth - xmintocopyintile;
+				rightpaddingthistile = xmaxplusone - inimagewidth;
+			} else
+				widthtocopy = xmaxplusone - xmintocopyintile;
+			widthtoreadinbytes = widthtocopy * bytesperpixel;
+			widthtowriteinbytes =
+			    (widthtocopy + rightpaddingthistile) * bytesperpixel;
 
-			if (TIFFReadTile(in, inbuf, xminoftile,
+			if (lengthtocopy > 0 && widthtocopy > 0 &&
+			    TIFFReadTile(in, inbuf, xminoftile,
 			    yminoftile, 0, 0) < 0) {
 				TIFFError(TIFFFileName(in),
 				    "Error, can't read tile at "
@@ -421,13 +473,18 @@ cpTiles2Strip(TIFF* in, void * ambiguous_out,
 
 			cpBufToBuf(bufp + colb,
 			    inbufrow + (xmintocopyintile-xminoftile) *
-				bytesperpixel, lengthtocopy,
-			    widthtocopyinbytes,
-			    outscanlinesizeinbytes - widthtocopyinbytes,
-			    intilewidthinbytes - widthtocopyinbytes);
-			colb += widthtocopyinbytes;
+				bytesperpixel,
+			    paddingbytes,
+			    lengthtocopy, bottompaddingthistile,
+			    widthtocopy, rightpaddingthistile,
+			    bytesperpixel,
+			    outscanlinesizeinbytes - widthtowriteinbytes,
+			    intilewidthinbytes - widthtoreadinbytes);
+			colb += (widthtocopy + rightpaddingthistile) *
+			    bytesperpixel;
 		}
-		bufp += outscanlinesizeinbytes * lengthtocopy;
+		bufp += outscanlinesizeinbytes *
+		    (lengthtocopy + bottompaddingthistile);
 	}
 
 	if (output_to_jpeg_rather_than_tiff) {
@@ -467,6 +524,7 @@ static int
 cpStrips2Strip(TIFF* in, void * ambiguous_out,
 	int output_to_jpeg_rather_than_tiff,
 	uint32 xmin, uint32 ymin, uint32 width, uint32 length,
+	uint32 rightpadding, uint32 bottompadding,
 	unsigned char * outbuf, uint32 * y_of_last_read_scanline,
 	uint32 inimagelength)
 {
@@ -501,7 +559,9 @@ cpStrips2Strip(TIFF* in, void * ambiguous_out,
 		return (EXIT_INSUFFICIENT_MEMORY);
 	}
 
-	/* When compression method doesn't support random access: */
+fprintf(stderr, "cpStrips2Strip warning: padding is yet to be implemented\n");
+
+	/* when compression method doesn't support random access: */
 	if (input_compression != COMPRESSION_NONE) {
 		uint32 y;
 
@@ -550,7 +610,8 @@ cpStrips2Strip(TIFF* in, void * ambiguous_out,
 
 		cpBufToBuf(bufp,
 		    inbufrow + xmintocopyinscanline * bytesperpixel,
-		    1, widthtocopyinbytes,
+		    NULL,
+		    1, 0, widthtocopyinbytes, 0, bytesperpixel,
 		    outscanlinesizeinbytes - widthtocopyinbytes,
 		    inwidthinbytes - widthtocopyinbytes);
 		bufp += outscanlinesizeinbytes;
@@ -658,7 +719,7 @@ makeMosaicFromTIFFFile(char * infilename)
 	uint32 hoverlap, voverlap;
 	uint32 hnpieces, vnpieces;
 	uint32 ndigitshtilenumber, ndigitsvtilenumber, x, y;
-	uint16 planarconfig, spp, bitspersample;
+	uint16 planarconfig, spp, bitspersample, sampleformat;
 	uint64 outmemorysize, ouroutmemorysize;
 	char * prefix;
 	unsigned char * outbuf = NULL;
@@ -679,6 +740,7 @@ makeMosaicFromTIFFFile(char * infilename)
 	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &inimagelength);
 	TIFFGetFieldDefaulted(in, TIFFTAG_SAMPLESPERPIXEL, &spp);
 	TIFFGetFieldDefaulted(in, TIFFTAG_BITSPERSAMPLE, &bitspersample);
+	TIFFGetFieldDefaulted(in, TIFFTAG_SAMPLEFORMAT, &sampleformat);
 
 	if (output_JPEG_files && bitspersample != 8 && spp != 3) {
 		TIFFError(TIFFFileName(in),
@@ -722,7 +784,7 @@ makeMosaicFromTIFFFile(char * infilename)
 	    (requestedpiecelength == 0 ||
 		inimagelength <= requestedpiecelength) &&
 	    (mosaicpiecesize == 0 || outmemorysize <= mosaicpiecesize)) {
-		/* Nothing to do */
+		/* nothing to do */
 		TIFFClose(in);
 		if (verbose) {
 			if (mosaicpiecesize)
@@ -762,22 +824,40 @@ makeMosaicFromTIFFFile(char * infilename)
 		    (output_JPEG_files &&
 		    (outwidth >= JPEG_MAX_DIMENSION ||
 		    outlength >= JPEG_MAX_DIMENSION) ) ) {
-			if (outlength > outwidth && outlength % 2 == 0 &&
+			if (outlength > outwidth && (outlength % 2 == 0 || paddinginy) &&
 			    requestedpiecelength == 0 &&
-			    (requestedpiecelengthdivisor == 0 ||
-			     (outlength/2) % requestedpiecelengthdivisor == 0))
+			    (requestedpiecelengthdivisor == 0 || paddinginy ||
+			    (outlength/2) % requestedpiecelengthdivisor == 0)) {
+				if (outlength % 2 != 0)
+					outlength++; /* use padding */
 				outlength /= 2;
-			else if (outwidth % 2 == 0 &&
+				if (requestedpiecelengthdivisor != 0 &&
+				    outlength % requestedpiecelengthdivisor != 0)
+					outlength += requestedpiecelengthdivisor -
+					    outlength % requestedpiecelengthdivisor; /* use padding */
+			} else if ((outwidth % 2 == 0 || paddinginx) &&
 			    requestedpiecewidth == 0 &&
-			    (requestedpiecewidthdivisor == 0 ||
-			    (outwidth/2) % requestedpiecewidthdivisor == 0))
+			    (requestedpiecewidthdivisor == 0 || paddinginx ||
+			    (outwidth/2) % requestedpiecewidthdivisor == 0)) {
+				if (outwidth % 2 != 0)
+					outwidth++; /* use padding */
 				outwidth /= 2;
-			else if (outlength % 2 == 0 &&
+				if (requestedpiecewidthdivisor != 0 &&
+				    outwidth % requestedpiecewidthdivisor != 0)
+				outwidth += requestedpiecewidthdivisor -
+				    outwidth % requestedpiecewidthdivisor; /* use padding */
+			} else if ((outlength % 2 == 0 || paddinginy) &&
 			    requestedpiecelength == 0 &&
-			    (requestedpiecelengthdivisor == 0 ||
-			     (outlength/2) % requestedpiecelengthdivisor == 0))
-				outlength /= 2;
-			else {
+			    (requestedpiecelengthdivisor == 0 || paddinginy ||
+			     (outlength/2) % requestedpiecelengthdivisor == 0)) {
+			     	if (outlength % 2 != 0)
+			     		outlength++; /* use padding */
+			     	outlength /= 2;
+  				if (requestedpiecelengthdivisor != 0 &&
+  				    outlength % requestedpiecelengthdivisor != 0)
+  					outlength += requestedpiecelengthdivisor -
+  					    outlength % requestedpiecelengthdivisor; /* use padding */
+			} else {
 				outwidth = 0;
 				outlength = 0;
 				break; /* can't divide any dimension by 2 */
@@ -803,15 +883,27 @@ makeMosaicFromTIFFFile(char * infilename)
 
 	outbuf= _TIFFmalloc(ouroutmemorysize);
 	while ( outbuf == NULL ) {
-		if (outlength > outwidth && outlength % 2 == 0 &&
-		    (requestedpiecelengthdivisor == 0 ||
-		     outlength % requestedpiecelengthdivisor == 0))
-			outlength /= 2;
-		else if (outwidth % 2 == 0 &&
-		    (requestedpiecewidthdivisor == 0 ||
-		    outwidth % requestedpiecewidthdivisor == 0))
+		if (outlength > outwidth && (outlength % 2 == 0 || paddinginy) &&
+		    (requestedpiecelengthdivisor == 0 || paddinginy ||
+		     outlength % requestedpiecelengthdivisor == 0)) {
+                        if (outlength % 2 != 0)
+                                outlength++; /* use padding */
+		        outlength /= 2;
+		        if (requestedpiecelengthdivisor != 0 &&
+	                    outlength % requestedpiecelengthdivisor != 0)
+	                        outlength += requestedpiecelengthdivisor -
+		                    outlength % requestedpiecelengthdivisor; /* use padding */
+		} else if ((outwidth % 2 == 0 || paddinginx ) &&
+		    (requestedpiecewidthdivisor == 0 || paddinginx ||
+		    outwidth % requestedpiecewidthdivisor == 0)) {
+                        if (outwidth % 2 != 0)
+                                outwidth++; /* use padding */
 			outwidth /= 2;
-		else
+			if (requestedpiecewidthdivisor != 0 &&
+			  outwidth % requestedpiecewidthdivisor != 0)
+		                outwidth += requestedpiecewidthdivisor -
+		                    outwidth % requestedpiecewidthdivisor; /* use padding */
+		} else
 			break; /* can't divide any dimension by 2 */
 
 		computeMaxPieceMemorySize(inimagewidth,
@@ -847,6 +939,54 @@ makeMosaicFromTIFFFile(char * infilename)
 			outmemorysize / 1048576.0, outmemorysize);
 	}
 
+	if ((paddinginx || paddinginy) && spp != numberpaddingvalues) {
+		fprintf(stderr, "File \"%s\": number of padding values"
+			" (%u) does not match number of samples per"
+			" pixel (%u).\n",
+			infilename, numberpaddingvalues, spp);
+		free(outbuf);
+		return EXIT_SYNTAX_ERROR;
+	}
+
+	uint16 bytesperpixel= (bitspersample + 7) / 8;
+	uint8 paddingbytes[bytesperpixel * spp];
+	uint16 s;
+	for (s = 0 ; s < spp ; s++) {
+		/* Here we should use sampleformat; instead, we assume 
+		  unsigned integer format uint8 or uint16 or... (rather 
+		  than e.g. int16 or float) */
+		switch (paddingvalues[s][0]) {
+			case 'M': {
+				uint16 b;
+				for (b = 0 ; b < bytesperpixel ; b++)
+					paddingbytes[s * bytesperpixel + b]=
+					    0xff;
+				break;
+			}
+			default: {
+				unsigned long long u= strtoull(
+				    paddingvalues[s], NULL, 10);
+				uint16 b;
+				/* Here we should care for endianness ;
+				we write paddingbytes bigendianly */
+				for (b = 0 ; b < bytesperpixel ; b++) {
+					paddingbytes[
+					    (s + 1) * bytesperpixel - 1 - b]=
+						u & 0xff;
+					u >>= 8;
+				}
+			}
+                }
+        }
+
+	/*for (s = 0 ; s < spp ; s++) {
+		fprintf(stderr, "[sample %u] %s", s, paddingvalues[s]);
+		uint16 b;
+		for (b = 0 ; b < bytesperpixel ; b++)
+			fprintf(stderr, " %d", paddingbytes[s * bytesperpixel + b]);
+		fprintf(stderr, "\n");
+	}*/
+
 	prefix = searchPrefixBeforeLastDot(infilename);
 
 	ndigitshtilenumber = searchNumberOfDigits(hnpieces);
@@ -863,11 +1003,12 @@ makeMosaicFromTIFFFile(char * infilename)
 		uint32 xwithleftoverlap = x - leftoverlap;
 
 		uint32 outwidthwithrightoverlap = outwidth + hoverlap;
+		uint32 amountofpaddingatright= 0;
 		uint32 xrightboundary = x + outwidthwithrightoverlap;
 		    /* equal to xwithleftoverlap + outwidth + 2*hoverlap */
 		assert(xrightboundary >= x); /* detect overflows */
 		/* At this point the computed right boundary of the
-		 * piece may lie outside the input image for two
+		 * piece may lie outside the input image for three
 		 * (mutually nonexclusive) reasons:
 		 * - The user requested a specific piece width which is
 		 * not a divisor of the input image width, and this
@@ -880,16 +1021,26 @@ makeMosaicFromTIFFFile(char * infilename)
 		 * image: x + outwidth + hoverlap > inimagewidth. The
 		 * right overlap must be restricted to
 		 * inimagewidth-x-outwidth.
-		 * In both cases, xrightboundary > inimagewidth and
+		 * - We use padding in x to achieve a constraint (e.g. 
+		 * piece width is a multiple of some integer number). This 
+		 * piece must not have a smaller width and we should 
+		 * complete it with padding values.
+		 * In cases 1 and 2, xrightboundary > inimagewidth and
 		 * outwidthwithrightoverlap must be taken equal to
 		 * inimagewidth-x; otherwise, it can be taken equal to
 		 * outwidth+hoverlap. */
-		if (xrightboundary > inimagewidth)
-			outwidthwithrightoverlap = inimagewidth - x;
+		if (xrightboundary > inimagewidth) {
+			if (paddinginx)
+				amountofpaddingatright=
+				    xrightboundary - inimagewidth;
+			else
+				outwidthwithrightoverlap = inimagewidth - x;
+		}
 		outwidthwithoverlap = leftoverlap + outwidthwithrightoverlap;
 
 		assert(xwithleftoverlap < inimagewidth); /* xwlol would be < 0 */
-		assert(xwithleftoverlap + outwidthwithoverlap <= inimagewidth);
+		assert(xwithleftoverlap + outwidthwithoverlap <=
+			inimagewidth + amountofpaddingatright);
 
 		for (y = 0 ; y < inimagelength ; y += outlength) {
 			char * outfilename;
@@ -901,17 +1052,27 @@ makeMosaicFromTIFFFile(char * infilename)
 			uint32 ywithtopoverlap = y - topoverlap;
 
 			uint32 outlengthwithbottomoverlap= outlength + voverlap;
+			uint32 amountofpaddingatbottom= 0;
 			uint32 ybottomboundary = y + outlengthwithbottomoverlap;
 			assert(ybottomboundary >= y); /* detect overflows */
-			if (ybottomboundary > inimagelength)
-				outlengthwithbottomoverlap = inimagelength - y;
+			if (ybottomboundary > inimagelength) {
+				if (paddinginy)
+					amountofpaddingatbottom =
+					    ybottomboundary - inimagelength;
+				else
+					outlengthwithbottomoverlap =
+					    inimagelength - y;
+			}
 			outlengthwithoverlap =
 				topoverlap + outlengthwithbottomoverlap;
 
 			assert(ywithtopoverlap < inimagelength);
 			 /* ywtol would be < 0 */
 			assert(ywithtopoverlap + outlengthwithoverlap <=
-				inimagelength);
+				inimagelength + amountofpaddingatbottom);
+
+			if (dryrun)
+				continue;
 
 			my_asprintf(&outfilename, "%s_i%0*uj%0*u%s",
 			    prefix, ndigitsvtilenumber,
@@ -985,6 +1146,10 @@ makeMosaicFromTIFFFile(char * infilename)
 					    ywithtopoverlap,
 					    outwidthwithoverlap,
 					    outlengthwithoverlap,
+					    amountofpaddingatright,
+					    amountofpaddingatbottom,
+					    paddingbytes,
+					    inimagewidth, inimagelength,
 					    outbuf))) ||
 				     (!TIFFIsTiled(in) &&
 				      !(error = cpStrips2Strip(in, 
@@ -993,6 +1158,8 @@ makeMosaicFromTIFFFile(char * infilename)
 					    ywithtopoverlap,
 					    outwidthwithoverlap,
 					    outlengthwithoverlap,
+					    amountofpaddingatright,
+					    amountofpaddingatbottom,
 					    outbuf,
 					    &y_of_last_read_scanline,
 					    inimagelength)))))
@@ -1018,6 +1185,10 @@ makeMosaicFromTIFFFile(char * infilename)
 					ywithtopoverlap,
 					outwidthwithoverlap,
 					outlengthwithoverlap,
+					amountofpaddingatright,
+					amountofpaddingatbottom,
+					paddingbytes,
+					inimagewidth, inimagelength,
 					outbuf))) ||
 				     (!TIFFIsTiled(in) &&
 				      !(error = cpStrips2Strip(in, out,
@@ -1025,6 +1196,8 @@ makeMosaicFromTIFFFile(char * infilename)
 					ywithtopoverlap,
 					outwidthwithoverlap,
 					outlengthwithoverlap,
+					amountofpaddingatright,
+					amountofpaddingatbottom,
 					outbuf,
 					&y_of_last_read_scanline,
 					inimagelength))))
@@ -1036,7 +1209,7 @@ makeMosaicFromTIFFFile(char * infilename)
 
 				TIFFClose(out);
 			}
-			if (error && !return_code) /* Error code = 1st error */
+			if (error && !return_code) /* error code = 1st error */
 				return_code = error;
 		} /* for y */
 	} /* for x */
@@ -1050,11 +1223,12 @@ makeMosaicFromTIFFFile(char * infilename)
 static void
 usage()
 {
-	fprintf(stderr, "tiffmakemosaic v" PACKAGE_VERSION " license GNU GPL v3 (c) 2012-2017 Christophe Deroulers\n\n");
+	fprintf(stderr, "tiffmakemosaic v" PACKAGE_VERSION " license GNU GPL v3 (c) 2012-2021 Christophe Deroulers\n\n");
 	fprintf(stderr, "Quote \"Deroulers et al., Diagnostic Pathology 2013, 8:92\" in your production\n       http://doi.org/10.1186/1746-1596-8-92\n\n");
 	fprintf(stderr, "Usage: tiffmakemosaic [options] file1.tif [file2.tif...]\n");
 	fprintf(stderr, " Produces a mosaic if needed, for each of the given files. A mosaic is a set of one-strip TIFF files of equal size, each of which has bounded size, which reproduce the original file if reassembled contiguously. Options:\n");
 	fprintf(stderr, " -v                verbose monitoring\n");
+	fprintf(stderr, " -y                dry run (do not write output file(s))\n");
 	fprintf(stderr, " -T                report TIFF errors/warnings on stderr (no dialog boxes)\n");
 	fprintf(stderr, " -M <size in MiB>  max. memory req. of each piece of the mosaic (default 1024);\n");
 	fprintf(stderr, "                   0 for no limit\n");
@@ -1065,6 +1239,12 @@ usage()
 	fprintf(stderr, "                   dimensions that satisfy -M and/or -m, divide the full image\n");
 	fprintf(stderr, "                   in equal pieces by powers of 2, and are close to each other)\n");
 	fprintf(stderr, " -O <overlap>[%%]   overlap of adjacent pieces in pixels [or percent] (default 0)\n");
+	fprintf(stderr, " -P[X][Y] #[,#...] pad image if necessary in direction x and/or y (default:\n");
+	fprintf(stderr, "                   both) to satisfy -M, -m or -g requirements (e.g. so that\n");
+	fprintf(stderr, "                   width is a multiple of larger a power of 2), adding to the\n");
+	fprintf(stderr, "                   right and/or to the bottom pixels of value # (if 1\n");
+	fprintf(stderr, "                   sample/pixel) or #,# (if 2 samples per pixels), and so on;\n");
+	fprintf(stderr, "                   M for # means maximum possible value (e.g. 255 for 8-bit)\n");
 	fprintf(stderr, " -j[#]             output JPEG files (with quality #, 0-100, default 75)\n");
 	fprintf(stderr, " -c none[:opts]    output TIFF files with no compression \n");
 	fprintf(stderr, " -c x[:opts]       output TIFF compressed with encoding x (jpeg, lzw, zip, ...)\n");
@@ -1082,7 +1262,7 @@ usage()
 
 
 static int
-processZIPOptions(char* cp)
+processZIPOptions(const char* cp)
 {
 	if ( (cp = strchr(cp, ':')) ) {
 		do {
@@ -1100,7 +1280,7 @@ processZIPOptions(char* cp)
 
 
 static int
-processG3Options(char* cp)
+processG3Options(const char* cp)
 {
 	if( (cp = strchr(cp, ':')) ) {
 		if (defg3opts == (uint32) -1)
@@ -1122,7 +1302,7 @@ processG3Options(char* cp)
 
 
 static int
-processCompressOptions(char* opt)
+processCompressOptions(const char* opt)
 {
 	if (strncasecmp(opt, "none", 4) == 0) {
 		/*char * cp = strchr(opt, ':');*/
@@ -1183,6 +1363,18 @@ processCompressOptions(char* opt)
 		defcompression = COMPRESSION_JBIG;
 	} else if (strncasecmp(opt, "sgilog", 6) == 0) {
 		defcompression = COMPRESSION_SGILOG;
+#ifdef HAVE_ZSTD
+  } else if (strncasecmp(opt, "zstd", 4) == 0) {
+		if (!processZIPOptions(opt))
+			return 0;
+    defcompression = COMPRESSION_ZSTD;
+#endif
+#ifdef HAVE_WEBP
+  } else if (strncasecmp(opt, "webp", 4) == 0) {
+		if (!processZIPOptions(opt))
+			return 0;
+    defcompression = COMPRESSION_WEBP;
+#endif
 	} else
 		return (0);
 
@@ -1218,6 +1410,70 @@ processPieceGeometryOptions(char* cp, uint32* width, uint32* length)
 }
 
 
+static int
+processPaddingValuesOption(char * cp)
+{
+	char * cp2 = cp;
+
+	/* first pass to count the values */
+	numberpaddingvalues= 0;
+	while (*cp == ' ')
+		cp++;
+	if (*cp != 'M' && (*cp < '0' || *cp > '9')) {
+		fprintf(stderr, "Incorrect padding value(s) argument to"
+		    " option -p: %s\n", cp);
+		return 0;
+	}
+	numberpaddingvalues= 1;
+	while (*cp != 0) {
+		if (*cp == ',') {
+			(numberpaddingvalues)++;
+			if (numberpaddingvalues == 0) /* overflow */
+				return 0;
+		}
+		cp++;
+	}
+
+	if ((paddingvalues =
+	    _TIFFmalloc(numberpaddingvalues * sizeof(*paddingvalues))) == NULL) {
+		perror("Insufficient memory for padding values ");
+		exit(EXIT_INSUFFICIENT_MEMORY);
+	}
+
+	/* second pass to store the values */
+	numberpaddingvalues= 0;
+	cp = cp2;
+	while (*cp != 0) {
+		while (*cp == ' ')
+			cp++;
+		cp2 = cp;
+		switch (*cp) {
+		        case 'M': cp++; break;
+		        default:
+		                strtoull(cp, &cp, 10);
+                		if (errno) {
+					fprintf(stderr, "Incorrect padding value in argument to option -p:"
+					    "%s\n", cp);
+                			return 0;
+                		}
+                }
+		paddingvalues[numberpaddingvalues] = cp2;
+		numberpaddingvalues++;
+		while (*cp != 0 && *cp == ' ')
+			cp++;
+		if (*cp != 0) {
+			if (*cp != ',') {
+				fprintf(stderr, "Unexpected char after padding value in argument to option -p:"
+				    "%s\n", cp);
+				return 0;
+			}
+			cp++;
+		}
+	}
+	return 1;
+}
+
+
 static void
 stderrErrorHandler(const char* module, const char* fmt, va_list ap)
 {
@@ -1248,6 +1504,8 @@ int main(int argc, char * argv[])
 
 		if (argv[arg][1] == 'v')
 			verbose = 1;
+		else if (argv[arg][1] == 'y')
+			dryrun++;
 		else if (argv[arg][1] == 'T') {
 			TIFFSetErrorHandler(stderrErrorHandler);
 			TIFFSetWarningHandler(stderrWarningHandler);
@@ -1312,6 +1570,35 @@ int main(int argc, char * argv[])
 				return EXIT_SYNTAX_ERROR;
 			}
 			arg++;
+		} else if (argv[arg][1] == 'P') {
+			size_t arglength;
+			int i = 2;
+
+			while (argv[arg][i] != 0) {
+				switch (argv[arg][i]) {
+					case 'X':
+						paddinginx= -1; break;
+					case 'Y':
+						paddinginx= -1; break;
+					default:
+						fprintf(stderr, "Expected x or y for direction of padding after -p, got \"%s\"\n",
+						    &(argv[arg][2]));
+						usage();
+						return EXIT_SYNTAX_ERROR;
+					}
+				i++;
+			}
+			if (paddinginx == 0 && paddinginy == 0) {
+		  		paddinginx= -1;
+				paddinginy= -1;
+			}
+
+			if (arg+1 >= argc ||
+			    !processPaddingValuesOption(argv[arg+1])) {
+				usage();
+				return EXIT_SYNTAX_ERROR;
+			}
+			arg++;
 		} else if (argv[arg][1] == 'O') {
 			size_t arglength;
 
@@ -1381,6 +1668,9 @@ int main(int argc, char * argv[])
 	}
 
 	if (verbose) {
+		if (dryrun)
+			fprintf(stderr, "Dry run --- simulate only, compute"
+				" tile dimensions but don't write files.\n");
 		if (mosaicpiecesize)
 			fprintf(stderr, "Maximum memory requirement of each "
 			"produced piece: %.3f MiB (" UINT64_FORMAT
