@@ -1,6 +1,6 @@
 /* tiffsplittiles
 
- v. 1.3.10
+ v. 1.4
 
  Copyright (c) 2012-2017 Christophe Deroulers
 
